#!/usr/bin/env python
#
# Copyright (c) 2016-2020, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

import os
import sys
import shutil
import syslog
import subprocess
import json
import glob
import re
import pwd
import grp

from tempfile import mkdtemp
from infra_cep_helper import CEPCollecting

BACKEND_DIR = '/var/cache/cep'
MAX_REPORT_SIZE = 20*1024*1024 # 20Mb
RESULTS_DIR = '/vz/tmp/cep/output/final'
NGINX_LOG = '/var/log/nginx/access.log'


class ManagePy(CEPCollecting):

    def collect_backend_output(self, pre_cmd=""):
        home_folder = os.environ['HOME'] = "/usr/libexec/vstorage-ui-backend"
        ld_path = os.environ['LD_LIBRARY_PATH'] = "/opt/rh/rh-python36/root/usr/lib64:/opt/rh/rh-python34/root/usr/lib64 "
        user = os.environ['USER'] = "vstoradmin "
        new_path = os.environ['PATH'] = "/usr/libexec/vstorage-ui-backend/venv/bin:/opt/rh/rh-python36/root/usr/lib64:/opt/rh/rh-python34/root/usr/bin:/usr/local/bin:/bin:/usr/bin:/usr/local/sbin:/usr/sbin "
        backend_conf = 'UI_BACKEND_CONFIG=/usr/libexec/vstorage-ui-backend/etc/backend.cfg '
        venv = os.environ['VIRTUAL_ENV'] = "/opt/rh/rh-python36/root/usr/"

        cmd = pre_cmd + 'sudo -E -u ' + user + backend_conf + 'PATH=' + new_path + 'LD_LIBRARY_PATH='+ ld_path + venv + 'bin/python ' + home_folder  + '/manage.py make_report'
        if self.is_hci():
            cmd += ' --cep'

        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = proc.communicate()
        if proc.returncode != 0:
            print("Command [%s] reported an error" % cmd)
            sys.exit(1)

        # Obfuscate cluster name
        # Makes sense to move this logic to ui-agent or even vstorage-core scripts;
        # on the other hand, we drop cluster name in CEP but may want to preserve it
        # in other reports
        idx = 0
        for rep_file in glob.glob('{0}/cluster*report*'.format(BACKEND_DIR)):
            d = mkdtemp()
            idx += 1
            os.chdir(d)
            subprocess.check_output(["tar", "xf", rep_file])
            for rep_dir in glob.glob('{0}/vstorage-report*/vstorage_report.tgz'.format(d)):
                d2 = mkdtemp()
                os.chdir(d2)
                subprocess.check_output(["tar", "xf", rep_dir])
                subprocess.check_output(["sed", "-i", 's#<cluster_name>.*</cluster_name>#<cluster_name>clust{0}</cluster_name>#'.format(idx), "stat.xml"])
                subprocess.check_output(["tar", "czf", "vstorage_report.tgz"] + glob.glob("*"))
                os.unlink(rep_dir)
                shutil.move(d2 + "/vstorage_report.tgz", rep_dir)
                shutil.rmtree(d2)
            os.chdir(d)
            clust_report = "cluster_clust{0}_report.tgz".format(idx)
            subprocess.check_output(["tar", "czf", clust_report] + glob.glob("*"))
            if os.path.exists(RESULTS_DIR + "/" + clust_report):
                os.unlink(RESULTS_DIR + "/" + clust_report)
            shutil.move(d + "/" + clust_report, RESULTS_DIR)
            shutil.rmtree(d)

        if pre_cmd:
            with open(dstdir + '/cluster_report_backend.tar.gz', "w") as rep_file:
                proc = subprocess.Popen(pre_cmd + 'cat /tmp/backend_report.tar.gz', shell=True, stdout=rep_file, stderr=subprocess.PIPE)
                err = proc.communicate()

    def change_config(self, protocol):
        try:
            disp_helper_config = '/etc/vz/disp_helper.json'
            f = open(disp_helper_config, "r")
        except IOError:
            print("Failed to edit configuration file of the disp-helper daemon")
            sys.exit(2)
        try:
            json_data = json.load(f)
        except:
            f.close()
            return

        f.close()
        if json_data["protocol"] != protocol:
            # change protocol to what we want
            json_data["protocol"] =  "%s" % (protocol)
            try:
                with open (disp_helper_config, 'w') as f_new:
                    # set separators to prevent trailing spaces
                    json.dump(json_data, f_new, indent=2, separators=(',', ':'))
                self.service_restart_reload('restart', 'disp-helper')
            except:
                pass

    def backend_job(self):
        if self.service_active('vstorage-ui-backend'):
            # set protocol to CURL
            self.change_config('CURL')
            # reload service
        else:
            self.change_config('FILE')


    def web_language(self):
        lang_list = []
        lang_pattern = 'ext-lang(.*)'
        if self.service_active('vstorage-ui-backend'):
            if os.path.exists(NGINX_LOG) and os.path.getsize(NGINX_LOG) > 0:
                for i, line in enumerate(open(NGINX_LOG)):
                    for match in re.finditer(lang_pattern, line):
                        only_language = match.group().split()
                        try:
                            ext_lang = only_language[0].split('.', 1)[0]
                            if not ext_lang in lang_list:
                                lang_list.append(ext_lang)
                        except:
                            pass
        return list(set(lang_list))


    def openstack_users(self):
        openstack_dict = dict()

        SOURCE = '/etc/kolla/admin-openrc.sh'

        if not os.path.exists(SOURCE):
            return {}

        proc = subprocess.Popen(['bash', '-c', 'source {} && env'.format(SOURCE)], stdout=subprocess.PIPE)
        source_env = {tup[0].strip(): tup[1].strip() for tup in map(lambda s: s.strip().split('=', 1), proc.stdout)}
        command_users = '/usr/bin/openstack user list --insecure'
        users_count = subprocess.Popen(command_users, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=source_env)
        (out, err) = users_count.communicate()
        strings = out.count('\n')
        users = int(strings) - 4
        command_roles = '/usr/bin/openstack role assignment list --insecure --names'
        roles_count = subprocess.Popen(command_roles, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=source_env)
        (out, err) = roles_count.communicate()
        strings_roles = out.count('\n')
        roles = int(strings_roles) - 4
        category_match = re.findall('(\Wadmin[^@])', out)
        category_match_s = re.findall('(\Wsuperuser[^@])', out)

        openstack_dict["total_users"] = users
        openstack_dict["total_roles"] = roles
        openstack_dict["users_in_admin_role"] = len(category_match)
        openstack_dict["users_in_superuser_role"] = len(category_match_s)
        return openstack_dict

    def fill_results(self):
        if self.is_hci():
            self.backend_job()
            vstorage_backend = '/etc/kolla/vstorage-backend.py'
            if os.path.exists(vstorage_backend):
                # make /etc/kolla/admin-openrc.sh
                try:
                    self.execute_shell('kolla-ansible -i /etc/kolla/vstorage-backend.py post-deploy')
                    self.result["openstack_user_counts"] = self.openstack_users()
                except:
                    pass

        self.result["web_lang"] = self.web_language()

        # Try to prevent recursive re-packaging of reports due to changed master node
        if os.path.exists(BACKEND_DIR) and self.dir_size(BACKEND_DIR) > MAX_REPORT_SIZE:
            shutil.rmtree(BACKEND_DIR)

        try:
            os.makedirs(BACKEND_DIR)
        except:
            pass

        os.chown(BACKEND_DIR, pwd.getpwnam("vstoradmin").pw_uid, grp.getgrnam("vstoradmin").gr_gid)

        if self.service_active('vstorage-ui-backend'):
            o = self.collect_backend_output()

            net_report = '{0}/networks'.format(BACKEND_DIR)
            if os.path.exists(net_report):
                self.result["Compute"] = {}
                with open(net_report, 'r') as f:
                    networks = {}
                    net_id = None
                    for l in f.read().splitlines():
                        m = re.search(r'\A([\w\s]+?)\s{2,}(.+?)\Z', l)
                        if m:
                            k, v = m.group(1), m.group(2)
                            if k == 'ID':
                                net_id = v
                            elif k.find('Subnets') != -1:
                                net_id = None
                            elif net_id:
                                if net_id not in networks:
                                    networks[net_id] = {}
                                networks[net_id][k] = v
                    self.result["Compute"]["Networks"] = networks

            self.result["Vinfra"] = self.dump_vinfra()

            self.result["ABGWConf"] = self.read_ini_file("/etc/vstorage/abgw.config")
            for k in ['storage.s3.access-key-id', 'storage.s3.secret-access-key', 'storage.s3.bucket']:
                if k in self.result["ABGWConf"]:
                    del self.result["ABGWConf"][k]
        elif os.path.exists("/bin/vzlist"):
            try:
                proc = subprocess.Popen(["/bin/vzlist", "vstorage-ui"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                (out, err) = proc.communicate()
                if proc.returncode == 0:
                    o = self.collect_backend_output("/bin/vzctl exec vstorage-ui ")
            except Exception as e:
                syslog.syslog("exception in {}".format(e))
                pass
        else:
            print("this is not a master node, nothing to collect")
            sys.exit(1)


def main():
    o = ManagePy()
    o.run()


if __name__ == "__main__":
    main()
