#!/usr/bin/env python
#
# Copyright (c) 2016-2018, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

import time
import libvirt
import os
import re
from infra_cep_helper import CEPCollecting
from xml.etree import ElementTree
from xml.dom import minidom

class libvirt_get_info(CEPCollecting):
    def get_stats(self):
        r = dict()
        r["Total_amount_of_VMs"] = len(self.conn.listAllDomains())
        # active it's not "running" is vms in state_names
        r["Total_amount_of_active_VMs"] = len(self.conn.listDomainsID())
        # 0 domains if state_names return nothing
        r["Total_amount_of_inactive_VMs"] = len(self.conn.listDefinedDomains())

        if self.is_exist('docker') and self.service_active('docker') and self.container_exist('nova_compute'):
            r["CLUSTER_COMPUTE"] = "yes"

        host_nets = self.conn.listNetworks()
        if len ( host_nets ) > 0:
           for network in host_nets:
               r["host_network:" + str(network)] = self.get_info_about_networks(self.conn, network)

        active_hosts = self.conn.listDomainsID()
        if  len ( active_hosts ) > 0 :
            for id in active_hosts:
                r["DOMAIN_INFO_FOR_id_" + str(id)] = self.check_info_for_domain(self.conn, id)
        return r

    def createConnection(self):
        self.conn = libvirt.openReadOnly(None)
        if self.conn == None:
           print('Failed to open connection to QEMU/KVM')
           sys.exit(1)
        else:
            return self.conn

    def look_at_xml(self, dom, tag, element):
        raw_xml = dom.XMLDesc(0)
        xml = minidom.parseString(raw_xml)
        lst = []
        interfaceTypes = xml.getElementsByTagName(tag)
        for interfaceType in interfaceTypes:
            interfaceNodes = interfaceType.childNodes
            attrs = {}
            for interfaceNode in interfaceNodes:
                try:
                    if interfaceNode.nodeName[0:1] != '#':
                        if isinstance(element, str):
                            for k, v in interfaceNode.attributes.items():
                                if k == element:
                                    lst.append(v)
                                    break
                        else:
                            for k, v in interfaceNode.attributes.items():
                                if k in element:
                                    attrs[k] = v
                except:
                    pass
            if len(attrs.keys()) > 0:
                lst.append(attrs)
        return lst

    def get_info_about_networks(self, conn, network):
        net_dict = dict()
        network = conn.networkLookupByName(network)
        uuid = network.UUIDString()
        network = conn.networkLookupByUUIDString(uuid)
        net_dict["NETWORK_NAME"] = network.name()
        net_dict["NETWORK_UUID"] = network.UUIDString()
        net_dict["NETWORK_BRIDGE"] = network.bridgeName()
        net_dict["NETWORK_AUTOSTART"] = str(network.autostart())
        net_dict["NETWORK_ACTIVE"] = str(network.isActive())
        net_dict["NETWORK_PERSISTENT"] = str(network.isPersistent())

        return net_dict

    def get_block_stats(self, dom, disk_file):
        disk_stats_dict = dict()
        rd_req, rd_bytes, wr_req, wr_bytes, err = dom.blockStats(disk_file)
        disk_stats_dict['read requests issued:'] = str(rd_req)
        disk_stats_dict['bytes read:'] = str(rd_bytes)
        disk_stats_dict['write requests issued:'] = str(wr_req)
        disk_stats_dict['bytes written:'] = str(wr_bytes)
        disk_stats_dict['number of errors:'] = str(err)
        disk_stats_dict['disk file:'] = str(disk_file)

        return disk_stats_dict

    def get_network_stats(self, dom):
        net_stats_dict = dict()
        tree = ElementTree.fromstring(dom.XMLDesc())
        iface = tree.find('devices/interface/target').get('dev')
        stats = dom.interfaceStats(iface)
        net_stats_dict['interface'] = str(iface)
        net_stats_dict['rx bytes'] = str(stats[0])
        net_stats_dict['rx packets'] = str(stats[1])
        net_stats_dict['rx errors'] = str(stats[2])
        net_stats_dict['rx drops'] = str(stats[3])
        net_stats_dict['tx bytes'] = str(stats[4])
        net_stats_dict['tx packets'] = str(stats[5])
        net_stats_dict['tx errors'] = str(stats[6])
        net_stats_dict['tx drops'] = str(stats[7])

        return net_stats_dict

    def get_memory_stats(self, dom):
        mem_stats_dict = dict()
        stats  = dom.memoryStats()
        mem_stats_dict['memory used'] = stats['rss']
        mem_stats_dict['memory available'] = stats['actual']

        return mem_stats_dict

    def get_cpu_stats(self, dom):
        cpu_stats_dict = dict()
        stats = dom.getCPUStats(True)
        cpu_stats_dict['total cpu_time'] = str(stats[0]['cpu_time'])
        cpu_stats_dict['total system_time'] = str(stats[0]['system_time'])
        cpu_stats_dict['total user_time'] = str(stats[0]['user_time'])

        cpu_stats = dom.getCPUStats(False)
        for (i, cpu) in enumerate(cpu_stats):
           cpu_stats_dict['CPU '+str(i)] = str(cpu['cpu_time'] / 1000000000.)

        return cpu_stats_dict

    def template_info(self, machine_name):
        os_template = ''

        if os.path.isfile("/usr/bin/prlctl"):
            get_info = self.execute_shell('/usr/bin/prlctl list --info ' + machine_name)
            category_match = re.search('OS[^:]*:(.+)', str(get_info))
            if category_match:
                os_template = category_match.group(1).strip()

        elif os.path.isfile("/usr/bin/vzlist"):
            get_info = self.execute_shell('/usr/sbin/vzlist -a -H -o ostemplate ' + machine_name)
            if get_info:
                os_template = get_info.lstrip('.')

        return os_template

    def check_info_for_domain(self, conn, id):
        # digits to human-readable states
        state_names = {  libvirt.VIR_DOMAIN_NOSTATE     : "no state",
                         libvirt.VIR_DOMAIN_RUNNING     : "running",
                         libvirt.VIR_DOMAIN_BLOCKED     : "idle",
                         libvirt.VIR_DOMAIN_PAUSED      : "paused",
                         libvirt.VIR_DOMAIN_SHUTDOWN    : "in shutdown",
                         libvirt.VIR_DOMAIN_SHUTOFF     : "shut off",
                         libvirt.VIR_DOMAIN_CRASHED     : "crashed",
                         libvirt.VIR_DOMAIN_PMSUSPENDED : "suspended by PM" }

        vm_dict = dict()
        dom = self.conn.lookupByID(id)
        infos = dom.info()
        vm_dict["DOMAIN_NAME"] = dom.name()
        vm_dict["OS_TEMPLATE"] = self.template_info(dom.name())
        vm_dict["DOMAIN_STATE"] = state_names[infos[0]]
        vm_dict["DOMAIN_MAX_MEMORY"] = infos[1]
        vm_dict["DOMAIN_VIRT_CPUs"] = infos[3]
        model = conn.getSecurityModel()
        type = dom.OSType()
        vm_dict["SECURITY_MODEL"] = model
        vm_dict["OS_TYPE"] = type
        vm_dict["MEMORY STATS"] = self.get_memory_stats(dom)

        disk = self.look_at_xml(dom, 'disk', ['dev', 'file'])
        vm_dict["CPU_STATS"] = self.get_cpu_stats(dom)

        netdev = self.look_at_xml(dom, 'interface', ['dev', 'address'])
        netbr = self.look_at_xml(dom, 'interface', ['dev', 'bridge'])
        nettype = self.look_at_xml(dom, 'interface', ['dev', 'type'])

        vm_dict["ETHER_TYPE"] = self.look_at_xml(dom, 'interface', 'type')
        for el in netdev:
            vm_dict["MAC_ADDRESS_" + el['dev']] = el['address'] if 'dev' in el else ''
            vm_dict["NET_STATISTICS_" + el['dev']] = self.get_network_stats(dom)
        for el in netbr:
            vm_dict["BRIDGE_" + el['dev']] = el['bridge'] if 'dev' in el else ''
        for el in nettype:
            vm_dict["ETHER_TYPE_" + el['dev']] = el['type'] if 'dev' in el else ''
        for el in disk:
            vm_dict["DISK_DEVICE_" + el['dev']] = el['file'] if 'file' in el else ''
            disk_stats = self.get_block_stats(dom, el['file'] if 'file' in el else '')
            vm_dict["DISK_STATS_" + el['dev']] = disk_stats

        return vm_dict

    def fill_results(self):
        if not self.service_active('libvirtd'):
            return
        retries = 0
        succeed = False
        fail_reason = "Unknown reason"
        while not succeed and retries < 1:
            try:
                self.conn = libvirt.openReadOnly(None)
                if self.conn is None:
                    raise Exception("Unable to connect to libvirt")
                self.result["libvirtINFO"] = self.get_stats()
                succeed = True
            except Exception as e:
                retries += 1
                time.sleep(1 * 1 * retries)
                succeed = False
                fail_reason = str(e)
        if not succeed:
            raise Exception("Unable to collect libvirt stats info: " + fail_reason)

def main():
    o = libvirt_get_info()
    o.run()

if __name__ == "__main__":
    main()
