#!/usr/bin/env python
#
# Copyright (c) 2016-2018, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

import base64
import json
import time
import libvirt
import syslog
import traceback
from libvirt_qemu import qemuAgentCommand
from infra_cep_helper import CEPCollecting, MAX_SCRIPT_RETRIES

# Limit amount of data received from the guest per request
GUEST_OUT_DATA_LIMIT = 5000000

# Do not let commands inside guests loop forever
GUEST_CMD_TIMEOUT = 300

def exec_qga_command(domain, cmd):
    return json.loads(qemuAgentCommand(domain, json.dumps(cmd), 1, 0))


def exec_guest_command(domain, cmd, more_args=[]):
    try:
        a = cmd.split()
        s = {"execute": "guest-exec",
             "arguments": {"path": a[0],
                           "capture-output": True,
                           "arg": a[1:] + more_args}}
        status = exec_qga_command(domain, s)
        pid = status["return"]["pid"]
        spent_time = 0
        while True:
            s = {"execute": "guest-exec-status",
                 "arguments": {"pid": pid}}
            a = exec_qga_command(domain, s)
            if "out-data" in a["return"] and len(a["return"]["out-data"]) > GUEST_OUT_DATA_LIMIT:
                break
            if "exited" in a["return"] and not a["return"]["exited"]:
                spent_time += 1
                if spent_time >= GUEST_CMD_TIMEOUT:
                    syslog.syslog("command execution timed out")
                    break
                time.sleep(1)
                continue
            break

        if a["return"]["exitcode"] == 0:
            out_data = a["return"]["out-data"]
            if len(out_data) > GUEST_OUT_DATA_LIMIT:
                out_data = out_data[:GUEST_OUT_DATA_LIMIT]
                syslog.syslog("CEP output from guest %s has exceeded the output length limit" % domain.name())
            return base64.b64decode(out_data).strip().decode('utf-8')
    except:
        return ""


def exec_simple_qga_command(domain, cmd):
    s = {"execute": cmd}
    return exec_qga_command(domain, s)["return"]


def get_repos(domain, manager):
    repos_enabled = []
    yumlike = ['yum', 'dnf']
    deblike = ['apt']
    suselike = ['zypper']
    if manager in yumlike:
       check_repo = exec_guest_command(domain, "yum repolist enabled -q") or ""
       for line in check_repo.splitlines():
           if line.startswith("repo id"):
              continue
           a = ' '.join(line.split()[1:])
           repos_enabled.append(a)
    if manager in deblike:
       check_repo = exec_guest_command(domain, "apt-cache policy") or ""
       for line in check_repo.splitlines():
           if "http" in line:
              a = ' '.join(line.split()[1:])
              repos_enabled.append(a)
    if manager in suselike:
       check_repo = exec_guest_command(domain, "zypper -lr -E -e -") or ""
       for line in check_repo.splitlines():
           if line.startswith("baseurl"):
              a = ' '.join(line.split("=")[1:])
              repos_enabled.append(a)
           else:
              continue
    return repos_enabled


class LibvirtGuests(CEPCollecting):
    def __init__(self):
        self.fields = {
            "Fedora": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/usr/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                {"name": "Packages",
                 "cmd": "/usr/bin/rpm -qa"}
                ],
            "RHEL": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/usr/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                {"name": "Packages",
                 "cmd": "/usr/bin/rpm -qa"}
                ],
            "Debian": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                {"name": "Packages",
                 "cmd": "/usr/bin/dpkg-query --show"}
                ],
            "Ubuntu": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                {"name": "Packages",
                 "cmd": "/usr/bin/dpkg-query --show"}
                ],
            "CentOS": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/usr/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                {"name": "Packages",
                 "cmd": "/usr/bin/rpm -qa"}
                ],
            "VzLinux": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/usr/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                {"name": "Packages",
                 "cmd": "/usr/bin/rpm -qa"}
                ],
            "SUSE": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/usr/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                {"name": "Packages",
                 "cmd": "/usr/bin/rpm -qa"}
                ],
            "Linux": [
                {"name": "Services_systemd",
                 "cmd": "systemctl list-unit-files --no-legend"},
                {"name": "Services_sysvinit",
                 "cmd": "service --status-all"},
                {"name": "Processes",
                 "cmd": "/usr/bin/ps --no-headers -Ao \"\%U,\%t,\%a\""},
                ],
            "Windows": [
                {"name": "Services",
                 "cmd": "wmic service get "
                        "name,state,servicetype,startmode /format:csv"},
                {"name": "Processes",
                 "cmd": "wmic process get "
                        "caption,commandline,processid /format:csv"},
                {"name": "Packages",
                 "cmd": "wmic product get name,version /format:csv"}],
            "Unknown": []
        }

    def open_connect(self):
        self.conn = libvirt.open(None)
        if self.conn is None:
            raise Exception("Unable to connect to libvirt")

    def identify_os(self, domain):
        try:
            r = exec_guest_command(domain, "cmd \/c ver")
            if r.startswith("Microsoft Windows"):
                return {"type": "Windows", "version": r}
        except Exception:
            pass

        os_name = ""
        r = ""
        try:
            check_os = exec_guest_command(domain, "cat", ['/etc/os-release'])
            if not check_os:
               guess_platform_cmd = """import platform;d = platform.linux_distribution();print(d[0] + \":\" + d[1])"""
               r = exec_guest_command(domain, "/usr/bin/python -c", [guess_platform_cmd])
               if r:
                  (os_name, os_ver) = r.split(":")
               else:
                  r = exec_guest_command(domain, "/usr/bin/python3 -c", [guess_platform_cmd])
                  (os_name, os_ver) = r.split(":")
            for item in check_os.splitlines():
                if item.startswith('NAME='):
                    os_name = item.split('=')[1]
                if item.startswith('VERSION='):
                    os_ver = item.split('=')[1]
        except Exception:
            pass

        if os_name == "":
            try:
                r = exec_guest_command(domain, "/bin/cat", ['/etc/system-release'])
                matches = re.findall(r"(.*) release (\S+)", r)
                if len(matches) > 0:
                    os_name = matches[0][0]
                    os_ver = matches[0][0]
            except Exception:
                pass

        if os_name:
            os = {}
            if "Ubuntu" in os_name:
                os = {"type": "Ubuntu", "version": os_ver}
            elif "Red Hat" in os_name:
                os = {"type": "RHEL", "version": os_ver}
            elif "CentOS" in os_name:
                os = {"type": "CentOS", "version": os_ver}
            elif "debian" in os_name or "Debian" in os_name:
                os = {"type": "Debian", "version": os_ver}
            elif "SUSE" in os_name:
                os = {"type": "SUSE", "version": os_ver}
            elif "Fedora" in os_name:
                os = {"type": "Fedora", "version": os_ver}
            elif "Virtuozzo" in os_name:
                os = {"type": "VzLinux", "version": os_ver}

            if os:
                return os

        try:
            r = exec_guest_command(domain, "uname -a")
            if r.startswith("Linux"):
                os = {"type": "Linux", "version": "Unknown"}
            return os
        except:
            pass

        return {"type": "Unknown", "version": "Unknown"}

    def get_domain_info(self, domain):
        r = {"name": domain.name()}
        try:
            exec_simple_qga_command(domain, "guest-ping")
        except:
            r["qga-status"] = "Unavailable"
            return r
        r["Info"] = exec_simple_qga_command(domain, "guest-info")
        cmd = "guest-network-get-interfaces"
        r["NetworkInfo"] = exec_simple_qga_command(domain, cmd)
        r["Vcpus"] = exec_simple_qga_command(domain, "guest-get-vcpus")
        # TODO guest-get-fsinfo seems to be unstable
        # r["FSInfo"] = exec_simple_qga_command(domain, "guest-get-fsinfo")
        os = self.identify_os(domain)
        r["OS"] = os["type"]
        r["version"] = os["version"]
        # fetch enabled repos
        rh_types = ['vzlinux', 'rhel', 'centos', 'fedora']
        deb_types = ['ubuntu', 'debian']
        suse_types = ['suse', 'sles', 'opensuse']

        if r["OS"].lower() in rh_types:
           r["REPOS"] = get_repos(domain, 'yum')
        if r["OS"].lower() in deb_types:
           r["REPOS"] = get_repos(domain, 'apt')
        if r["OS"].lower() in suse_types:
           r["REPOS"] = get_repos(domain, 'zypper')

        for field in self.fields[os["type"]]:
            if field['name'] in ['Services_sysvinit', 'Services_systemd', 'Processes', 'Services']:
                continue

            try:
                field_value = exec_guest_command(domain, field["cmd"])
            except:
                field_value = ""
            r[field["name"]] = field_value
        return r

    def get_active_domains(self):
        t = libvirt.VIR_CONNECT_LIST_DOMAINS_ACTIVE
        return self.conn.listAllDomains(t)

    def fill_results(self):
        if not self.service_active('libvirtd'):
            return
        retries = 0
        succeed = False
        fail_reason = "Unknown reason"
        while not succeed and retries < MAX_SCRIPT_RETRIES:
            try:
                self.open_connect()
                for d in self.get_active_domains():
                    self.result[d.UUIDString()] = self.get_domain_info(d)
                succeed = True
            except Exception as e:
                retries += 1
                succeed = False
                fail_reason = str(e)
                if retries >= MAX_SCRIPT_RETRIES:
                    break
                time.sleep(60 * 2 * retries)
        if not succeed:
            raise StandardError("Unable to collect libvirt guests info: " + fail_reason)


def main():
    o = LibvirtGuests()
    o.run()


if __name__ == "__main__":
    main()
