#!/usr/bin/env python
#
# Copyright (c) 2016-2018, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

import re
import time
import json
import os
from infra_cep_helper import CEPCollecting, MAX_SCRIPT_RETRIES


class Networking(CEPCollecting):
    def get_ethtool_output(self, device):
        value = ""
        key = ""
        device_hw = {}

        try:
            out = self.execute_shell("/sbin/ethtool " + device)
            for l in out.splitlines():
                if l.startswith("Settings"):
                    continue
                if ":" in l:
                    if key:
                        device_hw[key] = value
                    l_array = l.strip().split(":")
                    key = l_array[0].strip()
                    value = ":".join(l_array[1:]).strip()
                else:
                    value += " " + l.strip()
            if key:
                device_hw[key] = value.strip()
        except:
            return {}
        return device_hw


    def fill_results(self):
        retries = 1
        succeed = False
        fail_reason = "Unknown reason"
        while not succeed and retries < MAX_SCRIPT_RETRIES:
            try:
                out = self.execute_shell("/usr/sbin/ip -o link")
                devices = []
                devices_hw = []
                regex = re.compile("\d+:\s+([^:]+):(.*)$")
                for l in out.splitlines():
                    (name, value) = re.findall(regex, l)[0]
                    if self.is_hci():
                        devices.insert(0, {name: self.hide_hwaddrs(value)})
                    else:
                        devices.insert(0, {name: value})
                    devices_hw.insert(0, {name: self.get_ethtool_output(name)})
                self.result["Devices"] = devices
                self.result["Devices_Hardware"] = devices_hw

                cmd = "/usr/sbin/iptables -L -n --line-numbers | grep '^[[:digit:]]' | wc -l"
                self.result["Iptables_num_rules"] = str(int(self.execute_shell(cmd)))

                if os.path.exists("/usr/bin/prlsrvctl"):
                    cmd = "/usr/bin/prlsrvctl net list --json"
                    out = self.execute_shell(cmd)
                    networks = json.loads(out)
                    for network in networks:
                        self.prepare_tags(network)
                        network["XmlLabel"] = "Network"
                    self.result["VirtualNetworks"] = networks
                else:
                    self.result["VirtualNetworks"] = {}
                succeed = True
            except Exception as e:
                retries += 1
                succeed = False
                fail_reason = str(e)
                if retries >= MAX_SCRIPT_RETRIES:
                    break
                time.sleep(60 * 2 * retries)
        if not succeed:
            raise Exception("Unable to collect networking info: " + fail_reason)


def main():
    o = Networking()
    o.run()


if __name__ == "__main__":
    main()
