#!/usr/bin/env python
#
# Copyright (c) 2016-2018, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#
try:
    import rpmUtils.transaction
    from rpmUtils.transaction import initReadOnlyTransaction
except:
    import dnf.rpm.transaction
    from dnf.rpm.transaction import initReadOnlyTransaction
from infra_cep_helper import CEPCollecting
import subprocess
import psutil


class Packages(CEPCollecting):
    def detect_processes(self, non_standard_pkgs):
        running_process = []
        weird_binaries = []
        pid_list = psutil.pids()
        for pid in pid_list:
            p = psutil.Process(pid)
            # sometimes happens really weird things
            # p.exe() returns /usr/sbin/iscsi-scstd;5d6e5289: for few files
            # let's trick it
            # append process executable to the list
            running_process.append(p.exe().split(';', 1)[0].strip(' '))
        # remove empty lines
        filtered = [x for x in running_process if len(x.strip()) > 0]
        for progname in filtered:
            try:
               check = subprocess.check_output(['/usr/bin/rpm', '-qf', progname, '--qf=%{NAME}-%{VERSION}-%{RELEASE}'])
               if any(check in pkg for pkg in non_standard_pkgs):
                   weird_binaries.append("file " + progname + " is owned by 3rd party package ")
            except subprocess.CalledProcessError as e:
               # exception here means process already exited
               weird_binaries.append(e.output)
        return weird_binaries

#    Not used - takes too much time
#    def check_consistency(self, pkg_name):
#        try:
#           check_sig = subprocess.check_output(['/usr/bin/rpm', '-V', pkg_name])
#        except subprocess.CalledProcessError:
#           return pkg_name

    def fill_results(self):
        # rpm -qa > rpm.list
        # for i in $(cat rpm.list);do rpm -q --qf "%{NAME}\n" $i;done
        if self.is_hci():
            standard_rpms = [line.rstrip('\n') for line in open('/usr/share/disp-helper/data/hci-rpm.list')]
        specific_packages = {
            "openvswitch": "OpenVSwitch",
            "openstack-nova-compute": "Nova",
            "openstack-cinder": "Cinder",
            "openstack-neutron": "Neutron",
            "zabbix": "Zabbix",
            "nagios": "Nagios",
            "shaman ": "shaman",
            "shaman2": "shaman",
            "cpupools-features": "cpupools",
            "pdrs ": "pdrs",
            "va-agent-core": "VA agent core",
            "bird": "BGP client",
            "extility-xvpagent": "FCO",
            "paci_hn_psbm": "PACI",
            "svmstack-nginx": "SolusVM"
        }
        rpm_packages = list()
        not_standard_hci = list()
        weird_packages = list()
        # VzLinux, Acronis and Virtuozzo signatures are ok
        vz_signatures = ['8d0ba01812f4d9', '0b7d1944cdad2a', 'bad1edb7f44187']
        specific_soft = list()
        unsigned_rpms = list()
        try:
            ts = initReadOnlyTransaction()
            mi = ts.dbMatch()
            for h in mi:
                if self.is_hci():
                    if h["name"] not in standard_rpms:
                       not_standard_hci.insert(0, "-".join([h["name"], h["version"], h["release"]]))
                rpm_packages.insert(0, "-".join([h["name"], h["version"], h["release"]]))
                # extract signature
                try:
                    sig = h.sprintf("%{SIGPGP:pgpsig}").split()[-1]
                    if sig:
                        if not any(signature in sig for signature in vz_signatures):
                            weird_packages.insert(0, 'signature: {} '.format(sig) + "-".join([h["name"], h["version"], h["release"]]))
                    if not sig:
                        unsigned_rpms.append('-'.join([h["name"], h['version'], h['release']]))
                except:
                    pass
                if h["name"] in specific_packages:
                    specific_soft.insert(0, specific_packages[h["name"]])
        except:
            pass
        # rpm -V take some time
#        for pkg in l:
#            bad_check.append(self.check_consistency(pkg.split()[-1]))
#        self.result['FAILED_VALIDATION'] = "\n".join(filter(None, bad_check))
        self.result["Packages"] = "\n".join(rpm_packages)
        if self.is_hci():
            self.result["NotStandartButFromHCI"] = "\n".join(not_standard_hci)
        self.result["Not_VZ_RPMS"] = "\n".join(weird_packages)
        self.result["SpecificSoftware"] = "\n".join(specific_soft)
        self.result["Unsigned_RPMS"] = "\n".join(unsigned_rpms)
        try:
            self.result["WeirdProcesses"] = "\n".join(self.detect_processes(weird_packages + unsigned_rpms))
        except:
            self.result["WeirdProcesses"] = ""


def main():
    o = Packages()
    o.run()


if __name__ == "__main__":
    main()
