#!/usr/bin/env python
#
# Copyright (c) 2016-2020, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

import os
import re
from datetime import datetime
from infra_cep_helper import CEPCollecting


class Static(CEPCollecting):
    def prepare_lsmod_info(self):
        out = self.execute_shell("/usr/sbin/lsmod")
        lines = out.splitlines()

        for l in lines:
            try:
                self.result["KernelModules"].insert(0, l.split(' ', 1)[0])
            except:
                pass

    def prepare_df_info(self):
        list_df = []
        out = self.execute_shell("/usr/bin/df")
        lines = out.splitlines()

        # ignore first line here
        for l in lines[1:]:
            try:
                list_df.append(l.split(' ', 1))
            except:
                pass
        self.result["DiskUsage"] = list_df

    def prepare_sysctl_info(self):
        out = self.execute_shell("/usr/sbin/sysctl -a")
        regex = re.compile("([^=\s]+)\s=\s(.*)")
        lines = out.splitlines()

      #  try:
        for l in lines:
            name = ''
            try:
                (name, value) = re.findall(regex, l)[0]
            except:
                pass
            if name:
                self.result["Sysctl"][name] = value

    def add_ip_info(self):
        self.result["ip_a"] = ""
        try:
            lines = self.execute_shell("/sbin/ip a").strip().split('\n')
            for l in lines:
                if 'link/ether' in l or 'inet' in l:
                    continue
                self.result["ip_a"] += l
        except:
            pass

    def add_top_info(self):
        try:
            self.result["top"] = self.execute_shell("/usr/bin/top -b -n1").strip()
        except:
            pass

    def add_virtwhat_info(self):
        vw = self.execute_shell("/usr/sbin/virt-what").splitlines()
        try:
            self.result["virt_what"] = " ".join(vw)
        except:
            pass

    def cloud_config_info(self):
        cloud_config = "/etc/cloud/cloud.cfg"
        # just ignore if not exist
        if not os.path.exists(cloud_config):
            return
        try:
            with open(cloud_config) as f:
                for l in f.read().splitlines():
                    # extract value. e.g. datasource_list: [ Ec2 ]
                    m = re.findall(r"datasource_list: {0,}\[(.*?)\]", l)
                    if m:
                        self.result["cloud_config"] = m[0].strip()
        except:
            pass

    def add_sestatus_info(self):
        try:
            self.result["SELinux"] = self.execute_shell("/usr/sbin/sestatus").split()[2]
        except:
            sysfs_selinux = "/sys/fs/selinux"
            if os.path.exists(sysfs_selinux):
                self.result["SELinux"] = "/usr/sbin/sestatus failed but /sys/fs/selinux exist!"
            else:
                self.result["SELinux"] = "disabled"

    def prepare_uname_info(self):
        try:
            self.result["Uname"] = self.execute_shell("/usr/bin/uname -a").strip()
        except:
            self.result["Uname"] = "unknown"

    def prepare_uptime_info(self):
        try:
            self.result["Uptime"] = self.execute_shell("/usr/bin/uptime").strip()
        except:
            self.result["Uptime"] = "unknown"

    def add_vz_conf(self):
        if not os.path.isfile("/etc/vz/vz.conf"):
            return
        self.result["VZConf"] = self.read_ini_file("/etc/vz/vz.conf")

    def add_pfcache_conf(self):
        d = dict()
        if not os.path.isfile("/etc/vz/pfcache.conf"):
            return

        regex = re.compile("^([^=\#]+)=([^\#]*)$")
        with open("/etc/vz/pfcache.conf", "r") as f:
            for l in f.readlines():
                try:
                    (name, value) = re.findall(regex, l.strip())[0]
                    d[name] = value.strip("\"")
                except:
                    continue
        self.result["Pfcache"] = d

    def add_vcmmd_conf(self):
        if not os.path.isfile("/etc/vz/vcmmd.conf"):
            return
        with open("/etc/vz/vcmmd.conf", "r") as f:
            self.result["VCMMDConf"] = f.read()

    def add_session_info(self):
        self.result["Emergency"] = {"OOMs": [], "Crashes": []}
        self.result["Reboots"] = list()
        self.result["Shutdowns"] = list()
        with open("/var/log/messages", "r") as f:
            for l in f:
                if "out of memory" in l:
                    self.result["Emergency"]["OOMs"].insert(0, l.strip())

        out = self.execute_shell("/usr/bin/last -x")
        for l in out.splitlines():
            if "crash" in l:
                self.result["Emergency"]["Crashes"].insert(0, l.strip())
            elif "reboot" in l:
                self.result["Reboots"].insert(0, l.strip())
            elif "shutdown" in l:
                self.result["Shutdowns"].insert(0, l.strip())

    def prepare_kpatch_info(self):
        sysfs_kpatch = "/sys/kernel/kpatch/patches"
        if not os.path.exists(sysfs_kpatch):
            sysfs_kpatch = "/sys/kernel/kpatch"
            if not os.path.exists(sysfs_kpatch):
                sysfs_kpatch = "/sys/kernel/livepatch"
        try:
            d = {"Installed": [], "Modules": []}
            for patch in os.listdir(sysfs_kpatch):
                with open(os.path.join(sysfs_kpatch, patch, "enabled")) as f:
                    s = {"Patch": patch, "enabled": f.read().split()}
                    d["Modules"].insert(0, s)
            for dir in os.listdir("/var/lib/kpatch"):
                s = {"Kernel": dir,
                     "Patches": [os.path.splitext(tmp)[0] for tmp
                                 in os.listdir(os.path.join("/var/lib/kpatch",
                                                            dir))]}
                d["Installed"].insert(0, s)
        except:
            d = "Unavailable"
        self.result["kpatch"] = d

    def prepare_tcache_info(self):
        sysfs_tcache = "/sys/kernel/debug/cleancache/succ_gets"
        # just ignore if not exist
        if not os.path.exists(sysfs_tcache):
           pass
        try:
            d = []
            with open(os.path.join(sysfs_tcache)) as f:
                s = f.read().split()
        # just ignore if can't read
        except IOError:
            pass
        self.result["t_cache"] = s

    def add_openvz_switch_info(self):
        UPGRADE_LOGS = {
            '/var/log/do-upgrade-vz7.log': 'OpenVZ2Vz7Switch',
            '/var/log/do-upgrade-vz8.log': 'OpenVZ2Vz8Switch',
            '/var/log/do-upgrade-vzlin-vz8.log': 'VzLin2Vz8Switch',
            '/var/log/do-upgrade-vzlin-vz7.log': 'VzLin2Vz7Switch',
            '/var/log/do-upgrade-openvz7.log': 'VzLin2OpenVz7Switch',
            '/var/log/do-upgrade-openvz8.log': 'VzLin2OpenVz8Switch'
        }

        for log in UPGRADE_LOGS:
            s = 0
            try:
                if os.path.isfile(log):
                    s = datetime.fromtimestamp(os.path.getmtime(log))
            except:
                pass
            self.result[UPGRADE_LOGS[log]] = str(s)

    def add_autoupdate_logs(self):
        p = "/var/log/vz_auto_update.log"
        if not os.path.isfile(p):
            return
        try:
            with open(p, "r") as f:
                self.result["AutoUpdateLog"] = f.read()
        except:
            pass

    def add_security_level_info(self):
        p = "/etc/vz/dispatcher.xml"
        if os.path.isfile(p):
            try:
                security_cmd = self.execute_shell('/usr/bin/prlsrvctl info')
                category_match = re.search('Minimal security level:(.+)', security_cmd)
                security_level = category_match.group(1).lstrip(' ')
                self.result["Minimal_security_level"] = str(security_level)
            except:
                pass

    def add_dispatcher_xml(self):
        if not os.path.isfile("/etc/vz/dispatcher.xml"):
            return
        backup_dir = None
        try:
            with open("/etc/vz/dispatcher.xml", "r") as f:
                self.result["dispatcher.xml"] = f.read()
                m = re.search(r'<DefaultBackupDirectory>(.+)</DefaultBackupDirectory>', result['dispatcher.xml'])
                if m:
                    backup_dir = m.groups(0)[0]
        except:
            pass

        if backup_dir:
            try:
                fd = os.open(backup_dir + "/check.txt", os.O_RDWR|os.O_CREAT|os.O_DIRECT)
                os.close(fd)
                os.unlink(backup_dir + "/check.txt")
                self.result["backup_odirect"] = 'Ok'
            except Exception as e:
                self.result["backup_odirect"] = str(e)
                pass
        else:
            self.result["backup_odirect"] = 'Unknown'


    def add_proc_snmp(self):
        if not os.path.isfile("/proc/net/snmp"):
            return
        try:
            with open("/proc/net/snmp", "r") as f:
                self.result["proc_snmp"] = f.read()
        except:
            pass

    def add_proc_netstat(self):
        if not os.path.isfile("/proc/net/snmp"):
            return
        try:
            with open("/proc/net/netstat", "r") as f:
                self.result["proc_netstat"] = f.read()
        except:
            pass

    def add_vz6_switch_info(self):
        p = "/var/log/vzupgrade.log"
        s = 0
        try:
            if os.path.isfile(p):
                s = datetime.fromtimestamp(os.path.getmtime(p))
        except:
            pass
        self.result["Vz6Switch"] = str(s)

    def add_guest_collector_info(self):
        try:
            is_executable = os.access("/usr/share/virtuozzo/cep-scripts/libvirt_guests.py", os.X_OK)
            self.result["CollectGuests"] = str(is_executable)
        except:
            self.result["CollectGuests"] = "False"

    def add_anaconda_details(self):
        d = {"GraphicInstaller": 0}
        if os.path.isfile("/var/log/anaconda/X.log"):
            d["GraphicInstaller"] = 1
        d["KernelParams"] = []
        p = "/var/log/anaconda/syslog"
        if os.path.isfile(p):
            regex = re.compile(".*kernel.*ommand line.*")
            with open(p, "r") as f:
                for line in f.readlines():
                    try:
                        if regex.match(line):
                            d["KernelParams"].append(line)
                    except:
                        if regex.match(line.decode('utf-8')):
                            d["KernelParams"].append(line.decode('utf-8'))
                    if len(d["KernelParams"]) == 2:
                        break
        self.result["Anaconda"] = d

    def add_vzplatform_details(self):
        try:
            with open("/etc/kolla/hosts", "r") as f:
                hosts = f.read()
                hosts = hosts.strip()[1:-1]
                hosts = hosts.split(',')
                for i, s in enumerate(hosts):
                    hosts[i] = s.strip()[1:-1]
                self.result["VzPlatformHosts"] = hosts
        except:
            pass

    def add_hwid(self):
        if not os.path.isfile("/usr/sbin/vz-get-hwid"):
            return
        try:
            hwid_out = self.execute_shell('/usr/sbin/vz-get-hwid')
            self.result["HWID"] = hwid_out.strip()
        except:
            pass

    def add_cpufeatures(self):
        try:
            out = self.execute_shell('/usr/sbin/cpufeatures info')
            self.result["cpufeatures_info"] = out.strip()
            out = self.execute_shell('/usr/sbin/cpufeatures dump --original')
            self.result["cpufeatures_dump"] = out.strip()
        except:
            pass

    def add_build_info(self):
        if os.path.exists("/etc/virtuozzo-release") or os.path.exists("/etc/openvz-release"):
            self.result["VirtuozzoBuild"] = {
                "Release": "Unknown",
                "Version": "Unknown"
            }
        else:
            self.result["ProductBuild"] = {
                "Release": "Unknown",
                "Version": "Unknown"
            }

        try:
            if os.path.exists("/etc/virtuozzo-release"):
                with open("/etc/virtuozzo-release", "r") as f:
                    build = f.read().replace('\n', '')
                    regex = re.compile("(.+)\s\((\d+)\)")
                    matches = re.findall(regex, build)
                    self.result["VirtuozzoBuild"]["Release"] = matches[0][0]
                    self.result["VirtuozzoBuild"]["Version"] = matches[0][1]
            elif os.path.exists("/etc/hci-release"):
                with open("/etc/hci-release", "r") as f:
                    build = f.read().replace('\n', '')
                    regex = re.compile("(.+)\s\((\d+)\)")
                    matches = re.findall(regex, build)
                    self.result["ProductBuild"]["Release"] = matches[0][0]
                    self.result["ProductBuild"]["Version"] = matches[0][1]
            elif os.path.exists("/etc/vzlinux-release"):
                with open("/etc/vzlinux-release", "r") as f:
                    build = f.read().replace('\n', '')
                    regex = re.compile("(.+)\s\((\d+)\)")
                    matches = re.findall(regex, build)
                    self.result["ProductBuild"]["Release"] = matches[0][0]
                    self.result["ProductBuild"]["Version"] = matches[0][1]
        except:
            pass


    def check_rsa_keys(self):
        root_rsa = "/root/.vz/keys/id_rsa"
        flag_status = "disabled"
        if os.path.exists(root_rsa):
           flag_status = "enabled"
        else:
           path = "/home/"
           rsa_files = [os.path.join(dp, f) for dp, dn, filenames in os.walk(path) for f in filenames]
           if any(".vz/keys/id_rsa" in string for string in rsa_files):
              flag_status = "enabled"
        self.result["prl_remote_auth"] = flag_status


    def check_leapp(self):
        leapp_dir = "/var/log/leapp"
        flag_status = False
        if os.path.exists(leapp_dir):
           flag_status = True
        self.result["Vz7Switch"] = flag_status


    def fill_results(self):
        self.result = {"KernelModules": [], "Sysctl": {}, "DiskUsage": {}}
        self.prepare_lsmod_info()
        self.prepare_df_info()
        self.prepare_sysctl_info()
        self.prepare_uname_info()
        self.prepare_uptime_info()
        try:
            self.add_vz_conf()
            self.add_vcmmd_conf()
            self.add_pfcache_conf()
            self.add_session_info()
        except:
            pass
        self.add_dispatcher_xml()
        self.add_proc_snmp()
        self.add_proc_netstat()
        self.prepare_kpatch_info()
        self.prepare_tcache_info()
        if not self.is_hci():
            self.add_security_level_info()
            self.add_openvz_switch_info()
            self.add_vz6_switch_info()
        self.add_cpufeatures()
        self.add_hwid()
        self.add_anaconda_details()
        self.add_vzplatform_details()
        self.add_build_info()
        self.add_ip_info()
        self.add_top_info()
        self.add_sestatus_info()
        self.cloud_config_info()
        self.add_virtwhat_info()
        self.add_guest_collector_info()
        self.check_rsa_keys()
        self.check_leapp()
        self.add_autoupdate_logs()


def main():
    o = Static()
    o.run()


if __name__ == "__main__":
    main()
