#!/usr/bin/env python
#
# Copyright (c) 2016-2018, Virtuozzo International GmbH
#
# This file is part of disp-helper. It's free software; you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
# Schaffhausen, Switzerland.
#

from infra_cep_helper import CEPCollecting
from lxml import etree
import os
import re
import json
import xmltodict
import subprocess
import hashlib

class Storage(CEPCollecting):
    def fill_results(self):
        l = []

        if not os.path.isdir("/etc/vstorage/clusters"):
            self.result['StorageClusters'] = l
            return

        idx = 0
        for cluster_name in os.listdir('/etc/vstorage/clusters'):
            cluster_stats = self.execute_shell("vstorage -c " + cluster_name + " stat -X").strip()
            root = etree.fromstring(cluster_stats)
            # We used to hide cluster name but then decided that it's ok to keep it
#            fake_cluster_name = "clust" + str(idx)
            fake_cluster_name = cluster_name
            idx += 1

            cluster_info = {"Name": fake_cluster_name,
                            "ID"     : root.find('./cluster_id').text,
                            "MDS"    : len(root.findall('./mds_list/mds')),
                            "CS"     : len(root.findall('./cs_list/cs')),
                            "Clients": len(root.findall('./clients_list/client')),
                            "Hosts"  : len(root.findall('./host_list/host_stat')),
                            "Repl_n" : root.find('./repl/norm').text,
                            "Repl_m" : root.find('./repl/max').text,
                            "Repl_l" : root.find('./repl/limit').text,
                            "Alloc"  : root.find('./space/allocatable').text,
                            "Eff"    : root.find('./space/effective_total').text,
                            "Alloc_r": root.find('./space/allocatable_raw').text,
                            "Total"  : root.find('./space/total').text,
                            "Total_r": root.find('./space/total_raw').text,
                            "Free"   : root.find('./space/free').text
                        }
            l.append(cluster_info)

            # run vstorage get-event and get-config only if cluster exist
            self.result["Vstorage_Config_" + fake_cluster_name] = self.get_cfg_output(cluster_name)
            self.result["Event_" + fake_cluster_name] = self.get_events_output(cluster_name)
            self.result["License_" + fake_cluster_name] = self.get_license_output(cluster_name)
            self.result["Stats_" + fake_cluster_name] = self.get_stats_output(cluster_name, fake_cluster_name)
            self.result["read_cache"] = self.read_cache_stats()

            # pull psql tables only if postgresql is running
            if not self.service_active('postgresql'):
                continue

            try:
                out = self.execute_shell('sudo -u vstoradmin /bin/psql -t -c \"select MAX(timestamp) from audit_log where type=\'CreateCluster\'\"')
                creation_date = out
            except:
                creation_date = ''
            self.result["Creation_Date_" + fake_cluster_name] = creation_date

            self.save_database_output(cluster_name, fake_cluster_name, 'abgws', 1, [ 'node_id', 'cluster_id', 'is_active', 'is_registered', 'storage_type' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'alerts', 1, [ 'id', 'type', 'cluster_id', 'node_id', 'object_id', 'severity', 'details', 'datetime' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'alua_iscsi_targets', 1, [ 'id', 'type', 'iqn' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'alua_targets', 1, [ 'id', 'node_id', 'group_id', 'type', 'state' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'alua_targets_groups', 1, [ 'id', 'name', 'cluster_id', 'type', 'is_running', 'is_obsolete', 'is_acl_enabled', 'acl', 'is_chap_enabled', 'chap_user', 'luns', 'state' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'alua_volumes', 1, [ 'id', 'cluster_id', 'name', 'serial', 'size', 'used_size', 'group_id', 'params', 'limits', 'tpgs' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'compute_nodes', 1, [ 'id', 'is_controller', 'is_compute' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'cses', 3, [ 'id', 'node_id', 'disk_id', 'tier', 'disk_name', 'status', 'disk_type' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'iface_roles', 1, [ 'name', 'port' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'iscsi_luns', 3, [ 'lunno', 'target_iqn', 'cluster_id', 'size', 'used', 'is_online', 'description', 'failure_domain', 'redundancy', 'tier' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'iscsi_targets', 2, [ 'iqn', 'cluster_id', 'node_id', 'state', 'portals', 'ownername', 'mut_ownername', 'prometheus' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'roles_sets', 1, [ 'id', 'name', 'roles' ])

            # partially private tables
            self.save_database_output(cluster_name, fake_cluster_name, 'nfs_nodes', 1, [ '_id', '_nfs_id', '_has_configd', '_fs_active', '_fs_total', '_os_active', '_os_total' ], [ '_private_ip_address' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'nfs_shares', 2, [ '_nfs_id', '_id', '_nfs_node_id', '_redundancy', '_failure_domain', '_tier', '_status', '_used_space', '_total_space', '_krb_auth_enabled', '_krb_keytab', '_prometheus' ], [ '_name', '_ip_address', '_path_prefix' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'nfs_exports', 3, [ '_nfs_id', '_nfs_share_id', '_name', '_access_type', '_security_types', '_clients', '_squash' ], [ '_name', '_path', '_anonymous_gid', '_anonymous_uid' ])

            self.save_database_output(cluster_name, fake_cluster_name, 's3_clusters', 2, [ 'id', 'storage_cluster_id', 'use_http', 'use_https', 'redundancy', 'failure_domain', 'tier' ], [ 'domain_name', 'cert', 'key' ])
            self.save_database_output(cluster_name, fake_cluster_name, 's3_nodes', 1, [ 'id', 's3_cluster_id', 'configd', 'active_nses_count', 'active_oses_count', 'active_s3gws_count', 'total_nses_count', 'total_oses_count', 'total_s3gws_count' ], [ 'storage_cluster_id', 'private_ip_address' ])

            self.save_database_output(cluster_name, fake_cluster_name, 'ifaces', 2, [ 'node_id', 'name', 'agent_roles', 'type', 'duplex', 'speed', 'rolesset_id' ], [ 'ip_interfaces' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'users', 1, [ 'username', 'password', 'description', 'external_id' ])
            self.save_database_output(cluster_name, fake_cluster_name, 'snmp_settings', 1, [ 'id', 'is_enabled', 'is_traps_enabled' ], [ 'trap_ip_addr', 'trap_tcp_port', 'trap_community' ])

        self.result["StorageClusters"] = l

    def get_cfg_output(self, cluster_name):
        try:
            out = self.execute_shell("/usr/bin/vstorage -c " + cluster_name + " get-config")
            res = {}
            for l in out.splitlines():
                arr = l.split('=')
                if len(arr)==2:
                    res[arr[0].strip()] = arr[1].strip()
            return res
        except:
            return {}

    def read_cache_stats(self):
        l = []
        cache_file = '/mnt/vstorage/.vstorage.info/read_cache'
        if os.path.exists(cache_file) and os.path.getsize(cache_file) > 0:
           with open(cache_file, "r") as file:
               for line in file:
                   if line.startswith('#'):
                      continue
                   l.append(' '.join(line.split()))
        return l

    def get_events_output(self, cluster_name):
        try:
            result = {}
            # regex
            # all what we need
            # (MDS (\w+)\: (.*?)+\s*(.*))
            # strip brackets reges
            # \[.*?\]
            out = self.execute_shell("/usr/bin/vstorage -c " + cluster_name + " get-event -we")
            # sometimes square brackets exist, sometimes not, strip it
            strip_square_brackets = '\[.*?\]'
            # it's a string with warning/errors/info/etc
            warning_regex = '(\w+ (\w+)\: (.*?)+\s*(.*))'
            stripped = re.sub(strip_square_brackets, '', out)
            date_stripped = re.findall(warning_regex, stripped)
            for wrn in date_stripped:
                if wrn in res:
                   res[wrn]+=1
                else:
                   res[wrn] = 1
            for i in res.keys():
                arr = i[0].split(': ')
                if len(arr)>=2:
                    result[arr[0]] = ': '.join(arr[1:]) + " x " + str(res[i])
            return result
        except:
            return {}

    def get_stats_output(self, cluster_name, fake_cluster_name):
        try:
            out = self.execute_shell("/usr/bin/vstorage -c " + cluster_name + " stat --xml")
            return xmltodict.parse(out.replace(cluster_name, fake_cluster_name))
        except:
            return {}

    def get_hash(self, s):
        h = hashlib.sha512(s.encode('utf-8')).hexdigest()
        return h[0:32]

    def save_database_output(self, cluster_name, fake_cluster_name, table, primary_cols, select_cols, hide_cols=[]):
        out = self.execute_shell('sudo -u vstoradmin /bin/psql -Ac \"select %s from %s\"' % (', '.join(select_cols), table))
        lines = out.replace(cluster_name, fake_cluster_name).splitlines()

        if len(lines)<2:
            return {}

        cols = lines.pop(0).split('|')

        res = {}
        for l in lines:
            arr = l.split("|")
            if len(arr)>1:
                values = {}
                for idx, val in enumerate(arr):
                    col_name = cols[idx]
                    if col_name not in hide_cols:
                        values[col_name] = val

                if primary_cols==1:
                    c1 = arr[0]
                    if cols[0] in hide_cols:
                        c1 = self.get_hash(c1)
                    res[c1] = values
                elif primary_cols==2:
                    c1, c2 = arr[0], arr[1]
                    if cols[0] in hide_cols:
                        c1 = self.get_hash(c1)
                    if cols[1] in hide_cols:
                        c2 = self.get_hash(c2)
                    res[c1] = {}
                    res[c1][c2] = values
                elif primary_cols==3:
                    c1, c2, c3 = arr[0], arr[1], arr[2]
                    if cols[0] in hide_cols:
                        c1 = self.get_hash(c1)
                    if cols[1] in hide_cols:
                        c2 = self.get_hash(c2)
                    if cols[2] in hide_cols:
                        c3 = self.get_hash(c3)

                    res[c1] = {}
                    res[c1][c2] = {}
                    res[c1][c2][c3] = values

        self.result['check_databases_' + fake_cluster_name + '_' + table] = res

    def get_license_output(self, cluster_name):
        try:
            out = self.execute_shell("/usr/bin/vstorage -c " + cluster_name + " view-license")

            res = {}
            for l in out.splitlines():
                arr = l.split('=')
                if len(arr)!=2:
                    arr = l.split(':')
                if len(arr)>=2:
                    key = arr[0].strip()
                    if key == 'owner_name' and self.is_hci():
                        continue
                    if key == 'License expiration date':
                        key = 'expiration'
                    res[key] = ':'.join(arr[1:]).strip().strip('"')

            out = self.execute_shell("/usr/bin/vstorage -c " + cluster_name + " view-license -a")
            for l in out.splitlines():
                if not l.startswith("key:"):
                    continue
                arr = l.split(':')
                if len(arr)>=2:
                    res['key_number'] = ':'.join(arr[1:])
                break

            return res
        except:
            return {}

def main():
    o = Storage()
    o.run()

if __name__ == "__main__":
    main()
